# -*- coding: utf-8 -*-
# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
# sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

# instead, make sure the package is installed when building docs


# -- Project information -----------------------------------------------------

from greedy import package_info
project = 'D-Wave Greedy'
copyright = '2019, D-Wave Systems Inc.'
author = package_info.__author__
version = package_info.__version__
release = package_info.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autosummary',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx.ext.ifconfig'
]

autosummary_generate = True

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
source_suffix = ['.rst', '.md']
source_parsers = {'.md': 'recommonmark.parser.CommonMarkParser'}

# The master toctree document.
master_doc = 'index'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

add_module_names = False

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store', 'sdk_index.rst']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True

modindex_common_prefix = ['greedy.']

doctest_global_setup = """
import dimod
import greedy

from unittest.mock import Mock
from dwave.system.testing import MockDWaveSampler
import dwave.system
dwave.system.DWaveSampler = Mock()
dwave.system.DWaveSampler.side_effect = MockDWaveSampler
"""

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
import sphinx_rtd_theme
html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

def setup(app):
    app.add_css_file('cookie_notice.css')
    app.add_js_file('cookie_notice.js')
    app.add_config_value('target', 'repo', 'env')


intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'networkx': ('https://networkx.github.io/documentation/stable/', None),
    'oceandocs': ('https://docs.ocean.dwavesys.com/en/stable/', None),
    'sysdocs_gettingstarted': ('https://docs.dwavesys.com/docs/latest/', None),
}



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np
from parameterized import parameterized, parameterized_class

import dimod
from dimod.testing.sampler import BQM_SUBCLASSES

from greedy.sampler import SteepestDescentSampler


BQM_CLASSES = [(cls, ) for cls in BQM_SUBCLASSES]


@parameterized_class([
    # vary large & sparse problem optimization via sampling params
    {"params": dict(large_sparse_opt=False)},
    {"params": dict(large_sparse_opt=True)},
])
@dimod.testing.load_sampler_bqm_tests(SteepestDescentSampler)
class TestSteepestDescentSampler(unittest.TestCase):

    def test_instantiation(self):
        """The sampler must conform to `dimod.Sampler` interface."""

        sampler = SteepestDescentSampler()
        dimod.testing.assert_sampler_api(sampler)

    @parameterized.expand(BQM_CLASSES)
    def test_edges(self, BQM):
        """The sampler correctly handles edge cases."""

        sampler = SteepestDescentSampler()

        # empty bqm
        ss = sampler.sample(BQM.empty('SPIN'), **self.params)
        self.assertEqual(ss.first.sample, {})

        # single-variable problem
        ss = sampler.sample(BQM.from_ising({'x': 1}, {}), **self.params)
        self.assertEqual(ss.first.sample, {'x': -1})

    @parameterized.expand(BQM_CLASSES)
    def test_validation(self, BQM):
        """Inputs are validated."""

        sampler = SteepestDescentSampler()
        empty = BQM.from_ising({}, {})
        bqm = BQM.from_ising({'x': 1}, {})

        with self.assertRaises(TypeError):
            sampler.sample(bqm, num_reads=2.3, **self.params)

        with self.assertRaises(ValueError):
            sampler.sample(bqm, num_reads=0, **self.params)

        with self.assertRaises(ValueError):
            sampler.sample(bqm, initial_states=(), **self.params)

        with self.assertRaises(TypeError):
            sampler.sample(bqm, seed=2.3, **self.params)

        with self.assertRaises(ValueError):
            sampler.sample(bqm, seed=-1, **self.params)

        with self.assertRaises(ValueError):
            sampler.sample(bqm, initial_states_generator='invalid', **self.params)

        init = dimod.SampleSet.from_samples({'y': 1}, vartype='SPIN', energy=0)
        with self.assertRaises(ValueError):
            sampler.sample(bqm, initial_states=init, **self.params)

        init = dimod.SampleSet.from_samples({0: 1}, vartype='SPIN', energy=0)
        with self.assertRaises(ValueError):
            sampler.sample(empty, initial_states=init, **self.params)

    def test_small_convex_ising(self):
        """The sampler must converge to a global minimum of a convex Ising problem."""

        # a convex section of hyperbolic paraboloid in the Ising space,
        # with global minimum at (-1,-1)
        h = {0: 2, 1: 2}
        J = {(0, 1): -1}

        ss = SteepestDescentSampler().sample_ising(h, J, **self.params)

        self.assertEqual(len(ss), 1)
        self.assertEqual(len(ss.variables), 2)
        self.assertEqual(ss.record.sample.shape, (1, 2))
        np.testing.assert_array_equal(ss.record.sample[0], [-1, -1])

    def test_small_convex_qubo(self):
        """The sampler must converge to a global minimum of a convex QUBO problem."""

        # a convex section of hyperbolic paraboloid in the QUBO space,
        # with global minimum at (0,0)
        Q = {(0, 0): 2, (1, 1): 2, (0, 1): -1}

        ss = SteepestDescentSampler().sample_qubo(Q, **self.params)

        self.assertEqual(len(ss), 1)
        self.assertEqual(len(ss.variables), 2)
        self.assertEqual(ss.record.sample.shape, (1, 2))
        np.testing.assert_array_equal(ss.record.sample[0], [0, 0])

    @parameterized.expand(BQM_CLASSES)
    def test_variable_relabeling(self, BQM):
        """The sampler must accept non-integer/sequential variable names."""

        # use a small convex bqm
        bqm = BQM.from_ising({'x': 2, 'y': 2}, {'xy': -1})

        ss = SteepestDescentSampler().sample(bqm, **self.params)

        self.assertSetEqual(set(ss.variables), set(bqm.variables))
        self.assertDictEqual(ss.first.sample, {'x': -1, 'y': -1})

    @parameterized.expand(BQM_CLASSES)
    def test_unsortable_labels(self, BQM):
        """The sampler must accept unorderable variable names."""

        # use a small convex bqm
        bqm = BQM.from_ising({0: 2, 'a': 2}, {(0, 'a'): -1})

        ss = SteepestDescentSampler().sample(bqm, **self.params)

        self.assertSetEqual(set(ss.variables), set(bqm.variables))
        self.assertEqual(ss.first.energy, -5)

    @parameterized.expand(BQM_CLASSES)
    def test_reproducible_convergence(self, BQM):
        """On a convex problem, all samples must correspond to the global minimum."""

        # a convex section of hyperbolic paraboloid in the Ising space,
        # with global minimum at (-1,-1)
        bqm = BQM.from_ising({0: 2, 1: 2}, {(0, 1): -1})
        num_samples = 100

        # each sample is derived from a random initial state
        ss = SteepestDescentSampler().sample(
            bqm, num_reads=num_samples, **self.params)

        expected = np.tile([-1, -1], (num_samples, 1))

        np.testing.assert_array_equal(ss.record.sample, expected)

    @parameterized.expand(BQM_CLASSES)
    def test_initial_states_randomization(self, BQM):
        """Assuming uniform RNG, samples must follow a bimodal distribution."""

        # use a simple centrally symmetric hyperbolic paraboloid with
        # two minima in Ising space: (-1, 1) and (1, -1)
        bqm = BQM.from_ising({}, {'xy': 1})

        num = 1000
        tol = 0.10

        ss = SteepestDescentSampler().sample(
            bqm, num_reads=num, **self.params
        ).aggregate()

        # sanity check: two minima
        self.assertEqual(len(ss), 2)

        # check uniform distribution, with some tolerance `tol`
        inf = num / len(ss) - num * tol
        sup = num / len(ss) + num * tol
        for record in ss.record:
            self.assertTrue(inf < record.num_occurrences < sup)

    @parameterized.expand(BQM_CLASSES)
    def test_initial_states(self, BQM):
        """The sampler must deterministically converge from the initial state(s)."""

        # use a simple centrally symmetric hyperbolic paraboloid with
        # two minima in Ising space: (-1, 1) and (1, -1)
        bqm = BQM.from_ising({}, {(0, 1): 1})

        initial_states = dimod.SampleSet.from_samples(
            {0: -1, 1: -1}, vartype='SPIN', energy=0)

        # move along 0-dimension from (-1, -1) and settle in a local minimum (1, -1)
        ss = SteepestDescentSampler().sample(
            bqm, initial_states=initial_states, **self.params)

        self.assertEqual(len(ss), 1)
        self.assertDictEqual(ss.first.sample, {0: 1, 1: -1})

        # repeat this for 1000 samples, with initial state tiled
        num_reads = 1000
        ss = SteepestDescentSampler().sample(
            bqm, initial_states=initial_states,
            initial_states_generator='tile', num_reads=num_reads, **self.params
        ).aggregate()

        self.assertEqual(len(ss), 1)
        self.assertDictEqual(ss.first.sample, {0: 1, 1: -1})
        self.assertEqual(ss.first.num_occurrences, num_reads)

    @parameterized.expand(BQM_CLASSES)
    def test_initial_states_generation_and_validation(self, BQM):
        """Initial states are properly validated/expanded with the state generator."""

        sampler = SteepestDescentSampler()
        bqm = BQM.from_ising({'x': 1}, {})

        # num_reads inferred from initial_states
        init = dimod.SampleSet.from_samples([{'x': 1}, {'x': -1}],
                                            vartype='SPIN', energy=0)
        ss = sampler.sample(bqm, initial_states=init, **self.params)
        self.assertEqual(len(ss), len(init))

        # reads truncated with `num_reads`
        ss = sampler.sample(bqm, initial_states=init, num_reads=1, **self.params)
        self.assertEqual(len(ss), 1)

        # init states tiled according to `num_reads`
        init = dimod.SampleSet.from_samples({'x': 1}, vartype='SPIN', energy=0)
        ss = sampler.sample(
            bqm, num_reads=10, initial_states=init,
            initial_states_generator='tile', **self.params)
        self.assertEqual(len(ss), 10)
        self.assertEqual(list(ss.aggregate().samples()), [{'x': -1}])

        # tiling fails
        with self.assertRaises(ValueError):
            sampler.sample(
                bqm, initial_states=None, initial_states_generator='tile',
                **self.params)

        # tiling truncates
        init = dimod.SampleSet.from_samples([{'x': 1}, {'x': -1}],
                                            vartype='SPIN', energy=0)
        ss = sampler.sample(
            bqm, num_reads=1, initial_states=init,
            initial_states_generator='tile', **self.params)
        self.assertEqual(len(ss), 1)

        # none generator works
        init = dimod.SampleSet.from_samples([{'x': 1}, {'x': -1}],
                                            vartype='SPIN', energy=0)
        ss = sampler.sample(
            bqm, num_reads=1, initial_states=init,
            initial_states_generator='none', **self.params)
        self.assertEqual(len(ss), 1)

        # none generator fails
        with self.assertRaises(ValueError):
            sampler.sample(
                bqm, num_reads=1, initial_states=None,
                initial_states_generator='none', **self.params)

    @parameterized.expand([
        (([-1, -1], 'ab'), ),
        ((np.array([-1, -1]), 'ab'), ),
        ((np.array([-1, -1], dtype=np.int8), 'ab'), ),
    ])
    def test_initial_states_sample_like(self, initial_states):
        """Samples-like is accepted for initial_states."""

        # global minimum at (-1,-1)
        bqm = dimod.BQM.from_ising({'a': 2, 'b': 2}, {'ab': -1})

        ss = SteepestDescentSampler().sample(bqm, initial_states=initial_states)

        # result is identical to initial state, with zero downhill moves
        np.testing.assert_array_equal(ss.record.sample, [[-1, -1]])
        np.testing.assert_array_equal(ss.record.num_steps, [0])



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx

import dimod
import greedy


class SteepestDescentSimple(object):
    params = [1, 1000, 1000000]
    param_names = ['num_reads']
    repeat = (3, 10, 60)

    def setup(self, num_reads):
        self.sampler = greedy.SteepestDescentSampler()
        self.h = {0: 2, 1: 2}
        self.J = {(0, 1): -1}

    def time_single_flip(self, num_reads):
        self.sampler.sample_ising(self.h, self.J, num_reads=num_reads, seed=0)


class SteepestDescentComplete(object):
    """Test steepest descent on complete graphs."""

    params = ([100, 1000, 2000], [1, 10])
    param_names = ['graph_size', 'num_reads']
    repeat = (3, 10, 60)
    timeout = 300

    def setup(self, graph_size, num_reads):
        self.sampler = greedy.SteepestDescentSampler()
        self.bqm = dimod.generators.random.ran_r(r=1, graph=graph_size, seed=0)

    def time_ran1(self, graph_size, num_reads):
        self.sampler.sample(self.bqm, num_reads=num_reads, seed=0)


class SteepestDescentSparse(object):
    """Test steepest descent on Erdős-Rényi sparse graphs with varying density."""

    params = ([100, 1000, 2000], [0.05, 0.1, 0.25, 0.5], [1, 10])
    param_names = ['graph_size', 'graph_density', 'num_reads']
    repeat = (3, 10, 60)
    timeout = 300

    def setup(self, graph_size, graph_density, num_reads):
        self.sampler = greedy.SteepestDescentSampler()
        self.graph = nx.fast_gnp_random_graph(n=graph_size, p=graph_density, seed=0)
        self.bqm = dimod.generators.random.ran_r(r=1, graph=self.graph, seed=0)

    def time_ran1(self, graph_size, graph_density, num_reads):
        self.sampler.sample(self.bqm, num_reads=num_reads, seed=0)


class SteepestDescentLargeSparse(object):
    """Test steepest descent on large and sparse Erdős-Rényi graphs.

    Note:
        Because of practical limitations on BQM size in `dimod`, we limit
        the number of edges in test problems below to ~5M.
    """

    params = ([2000, 10000, 15000], [0.01, 0.05], [1, 10])
    param_names = ['graph_size', 'graph_density', 'num_reads']
    repeat = (3, 10, 60)
    timeout = 300

    def setup(self, graph_size, graph_density, num_reads):
        self.sampler = greedy.SteepestDescentSampler()
        self.graph = nx.fast_gnp_random_graph(n=graph_size, p=graph_density, seed=0)
        self.bqm = dimod.generators.random.ran_r(r=1, graph=self.graph, seed=0)

    def time_ran1(self, graph_size, graph_density, num_reads):
        self.sampler.sample(self.bqm, num_reads=num_reads, seed=0, large_sparse_opt=True)


class SteepestDescentCompleteOverBQMTypes(object):
    """Test steepest descent on complete graphs."""

    params = ([dimod.AdjArrayBQM,
               dimod.AdjDictBQM,
               dimod.AdjMapBQM,
               dimod.AdjVectorBQM], [1, 10])
    param_names = ['bqm_type', 'num_reads']
    repeat = (3, 10, 60)
    timeout = 300
    graph_size = 1000

    def setup(self, bqm_type, num_reads):
        self.sampler = greedy.SteepestDescentSampler()
        self.bqm = dimod.generators.random.ran_r(
            r=1, graph=self.graph_size, cls=bqm_type, seed=0)

    def time_ran1(self, bqm_type, num_reads):
        self.sampler.sample(self.bqm, num_reads=num_reads, seed=0)



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import warnings

import greedy


class TestSteepestDescentSampler(unittest.TestCase):

    def test_import(self):
        from greedy import package_info, __version__

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            self.assertEqual(__version__, package_info.__version__)
            self.assertEqual(__version__, greedy.package_info.__version__)
            self.assertIn('__version__', dir(package_info))

    def test_attributes(self):
        pi = greedy.package_info

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)

            self.assertIsNotNone(pi.__package_name__)
            self.assertIsNotNone(pi.__version__)
            self.assertIsNotNone(pi.__author__)
            self.assertIsNotNone(pi.__author_email__)
            self.assertIsNotNone(pi.__description__)
            self.assertIsNotNone(pi.__url__)
            self.assertIsNotNone(pi.__license__)

    def test_deprecation(self):
        with self.assertWarns(DeprecationWarning):
            greedy.package_info.__version__



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import greedy


class SteepestGradientDescentCython(object):
    params = [1, 1000, 1000000]
    param_names = ['num_samples']
    repeat = (3, 10, 60)

    def setup(self, num_samples):
        self.linear_biases = [2, 2]
        self.coupler_starts = [0]
        self.coupler_ends = [1]
        self.coupler_weights = [-1]
        self.initial_states = np.tile(np.array([1, 1], dtype=np.int8), (num_samples, 1))

    def time_single_flip(self, num_samples):
        samples, energies, num_steps = greedy.descent.steepest_gradient_descent(
            num_samples, self.linear_biases, self.coupler_starts,
            self.coupler_ends, self.coupler_weights, self.initial_states)



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import numpy as np
import dimod

from greedy.descent import steepest_gradient_descent


class SteepestGradientDescentCython(unittest.TestCase):
    large_sparse_opt = False

    def test_steepest_gradient_descent_on_small_convex_ising(self):
        """The sampler must converge to a global minimum of a convex Ising problem."""

        # a convex section of hyperbolic paraboloid in the Ising space,
        # with global minimum at (-1,-1):
        #
        #   h = {0: 2, 1: 2}
        #   J = {(0, 1): -1}

        num_samples = 1
        linear_biases = [2, 2]
        coupler_starts, coupler_ends, coupler_weights = [0], [1], [-1]
        initial_states = np.array([[1, 1]], dtype=np.int8)

        samples, energies, num_steps = steepest_gradient_descent(
            num_samples, linear_biases, coupler_starts, coupler_ends,
            coupler_weights, initial_states, self.large_sparse_opt)

        self.assertEqual(samples.shape, (1, 2))
        np.testing.assert_array_equal(samples, [[-1, -1]])
        np.testing.assert_array_equal(energies, [-5])


class SteepestGradientDescentLargeSparseCython(SteepestGradientDescentCython):
    large_sparse_opt = True



# Copyright 2020 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from collections import abc

import numpy as np

import dimod
from dwave.system.testing import MockDWaveSampler

from greedy.composite import SteepestDescentComposite


@dimod.testing.load_sampler_bqm_tests(SteepestDescentComposite(dimod.ExactSolver()))
class TestSteepestDescendComposite(unittest.TestCase):
    def test_instantiation_smoketest(self):
        sampler = SteepestDescentComposite(MockDWaveSampler())

        dimod.testing.assert_sampler_api(sampler)

    def test_sample_ising(self):
        sampler = SteepestDescentComposite(MockDWaveSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        num_reads = 100

        response = sampler.sample_ising(h, J, num_reads=num_reads)

        self.assertEqual(len(response), num_reads)

        for sample in response.samples():
            self.assertIsInstance(sample, abc.Mapping)
            self.assertEqual(set(sample), set(h))

        for sample, energy in response.data(['sample', 'energy']):
            self.assertIsInstance(sample, abc.Mapping)
            self.assertEqual(set(sample), set(h))
            self.assertAlmostEqual(dimod.ising_energy(sample, h, J), energy)

    def test_convex(self):
        # a convex section of hyperbolic paraboloid in the Ising space,
        # with global minimum at (-1,-1)
        bqm = dimod.BQM.from_ising({0: 2, 1: 2}, {(0, 1): -1})
        ground = dimod.SampleSet.from_samples_bqm({0: -1, 1: -1}, bqm)

        sampler = SteepestDescentComposite(dimod.ExactSolver())
        sampleset = sampler.sample(bqm).aggregate()

        np.testing.assert_array_almost_equal(
            sampleset.record.sample, ground.record.sample)
